﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Drawing;
using System.Drawing.Drawing2D;

namespace Function_Editor
{
    public partial class Function
    {
        public Types Type { get; set; }

        public bool IsRange { get; set; }

        public float Exponent1 { get; set; }
        public float Exponent2 { get; set; }

        public int ExponentType1 { get; set; }
        public int ExponentType2 { get; set; }

        public OutputTypes OutputType
        {
            get { return outputType; }
            set
            {
                outputType = value;
            }
        }
        private OutputTypes outputType;

        public PointF YBounds { get; set; }

        public Color[] Colors { get { return colors; } }
        private Color[] colors = new Color[4];

        public PointF[] Points1 { get; set; }
        public PointF[] Points2 { get; set; }

        public PointF FunctionBounds1 { get; set; }
        public PointF FunctionBounds2 { get; set; }

        public float Frequency1 { get; set; }
        public float Frequency2 { get; set; }

        public float Phase1 { get; set; }
        public float Phase2 { get; set; }

        #region Read
        public void ReadFrom(BinaryReader br)
        {
            Type = (Function.Types)br.ReadByte();
            byte data = br.ReadByte();
            IsRange = ((data & 1) == 1);
            ExponentType1 = br.ReadByte();
            ExponentType2 = br.ReadByte();

            outputType = (Function.OutputTypes)((int)data >> 4);

            if (OutputType == Function.OutputTypes.Scalar)
                ReadYBounds(br);
            else
                ReadColors(br);

            int headerEnd = (int)br.BaseStream.Position;

            switch (Type)
            {
                case Types.Identity:
                case Types.Constant:
                    break;
                case Types.Transition:
                case Types.Periodic:
                {
                    Frequency1 = br.ReadSingle();
                    Phase1 = br.ReadSingle();
                    FunctionBounds1 = new PointF(br.ReadSingle(), br.ReadSingle());

                    if (IsRange)
                    {
                        Frequency2 = br.ReadSingle();
                        Phase2 = br.ReadSingle();
                        FunctionBounds2 = new PointF(br.ReadSingle(), br.ReadSingle());
                    }
                    break;
                }
                case Types.Linear:
                case Types.LinearKey:
                //case Types.MultiLinearKey:
                case Types.Spline:
                //case Types.MultiSpline:
                case Types.Exponent:
                case Types.Spline2:
                {
                    Points1 = ReadPoints(br);

                    if (IsRange)
                    {
                        br.BaseStream.Position = headerEnd + 4 * Function.TypeSizeTable[(int)Type];
                        Points2 = ReadPoints(br);
                    }
                    break;
                }
            }
        }

        private void ReadYBounds(BinaryReader br)
        {
            YBounds = new PointF(br.ReadSingle(), br.ReadSingle());
            br.BaseStream.Position += 8;
        }

        private void ReadColors(BinaryReader br)
        {
            int colorCount = 4;
            if ((int)OutputType < 2)
                colorCount = (int)OutputType;
            for (int i = 0; i < colorCount; i++)
            {
                byte[] bytes = br.ReadBytes(4);
                Colors[i] = Color.FromArgb(255, bytes[2], bytes[1], bytes[0]);
            }
        }

        private PointF[] ReadPoints(BinaryReader br)
        {
            PointF[] points = new PointF[Function.GetNumberOfPoints(Type)];
            for (int i = 0; i < points.Length; i++)
                points[i] = new PointF(br.ReadSingle(), br.ReadSingle());
            return points;
        }
        #endregion

        #region Write
        public void WriteTo(BinaryWriter bw)
        {
            bw.Write((byte)Type);
            bw.Write((byte)((int)((IsRange) ? 1 : 0) | ((int)outputType << 4)));
            bw.Write((byte)ExponentType1);
            bw.Write((byte)ExponentType2);

            if (OutputType == Function.OutputTypes.Scalar)
                WriteYBounds(bw);
            else
                WriteColors(bw);

            int headerEnd = (int)bw.BaseStream.Position;

            switch (Type)
            {
                case Types.Identity:
                case Types.Constant:
                    break;
                case Types.Transition:
                case Types.Periodic:
                {
                    bw.Write(Frequency1);
                    bw.Write(Phase1);
                    bw.Write(FunctionBounds1.X);
                    bw.Write(FunctionBounds1.Y);

                    if (IsRange)
                    {
                        bw.Write(Frequency2);
                        bw.Write(Phase2);
                        bw.Write(FunctionBounds2.X);
                        bw.Write(FunctionBounds2.Y);
                    }
                    break;
                }
                case Types.Linear:
                case Types.LinearKey:
                //case Types.MultiLinearKey:
                case Types.Spline:
                //case Types.MultiSpline:
                case Types.Exponent:
                case Types.Spline2:
                {
                    WritePoints(bw, Points1);

                    if (IsRange)
                    {
                        bw.BaseStream.Position = headerEnd + 4 * Function.TypeSizeTable[(int)Type];
                        WritePoints(bw, Points2);
                    }
                    break;
                }
            }
        }

        private void WriteYBounds(BinaryWriter bw)
        {
            bw.Write(YBounds.X);
            bw.Write(YBounds.Y);
            bw.BaseStream.Position += 8;
        }

        private void WriteColors(BinaryWriter bw)
        {
            int colorCount = 4;
            if ((int)OutputType < 2)
                colorCount = (int)OutputType;
            for (int i = 0; i < colorCount; i++)
            {
                bw.Write(Colors[i].B);
                bw.Write(Colors[i].G);
                bw.Write(Colors[i].R);
                bw.Write((byte)0);
            }
        }

        private void WritePoints(BinaryWriter bw, PointF[] points)
        {
            for (int i = 0; i < points.Length; i++)
            {
                bw.Write(points[i].X);
                bw.Write(points[i].Y);
            }
        }
        #endregion
    }
}
