﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using H2.Streams;
using Function_Editor;
using System.IO;

namespace Squiggle
{
    public class TagBlock
    {
        public List<FunctionInfo> Functions;

        public class FunctionInfo
        {
            public string Name;
            public int PointerOffset;
            public int StartOffset;
            public int DataLength;

            public FunctionInfo(MapStream map, string name, int pointerOffset)
            {
                Name = name;
                PointerOffset = pointerOffset;
                DataLength = map.ReadReflexiveAt(pointerOffset);
                StartOffset = (int)map.Position;
            }

            public void Load(MapStream map, Function f)
            {
                f.ReadFrom(new BinaryReader(new MemoryStream(map.ReadBytesAt(StartOffset, DataLength, false))));
            }

            public void Save(MapStream map, Function f)
            {
                byte[] data = new byte[DataLength];
                f.WriteTo(new BinaryWriter(new MemoryStream(data)));
                map.WriteAt(StartOffset, data, false);
            }

            public override string ToString()
            {
                return Name;
            }
        }

        public static TagBlock Read(MapStream map, string tagType, string tagName)
        {
            if (tagName == null) return null;
            map.MoveToTag(tagType, tagName);
            switch (tagType)
            {
                case "bipd":
                case "bloc":
                case "crea":
                case "ctrl":
                case "eqip":
                case "garb":
                case "mach":
                case "proj":
                case "scen":
                case "vehi":
                case "weap":
                    return ReadObj(map);
                case "effe":
                    return ReadEFFE(map);
                case "jpt!":
                    return ReadJPT(map);
                case "lens":
                    return ReadLENS(map);
                case "ligh":
                    return ReadLIGH(map);
                case "MGS2":
                    return ReadMGS2(map);
                case "prt3":
                    return ReadPrt3(map);
                case "PRTM":
                    return ReadPRTM(map);
                case "tdtl":
                    return ReadTDTL(map);
            }
            return null;
        }

        public static TagBlock ReadObj(MapStream map)
        {
            int functionCount = map.ReadReflexiveAt((int)map.Position + 100);
            int StartOffset = (int)map.Position;

            TagBlock tb = new TagBlock();
            tb.Functions = new List<FunctionInfo>();
            for (int i = 0; i < functionCount; i++)
            {
                tb.Functions.Add(new FunctionInfo(
                    map,
                    string.Format("Functions[{0}]: {1} -> {2}", i, map.ReadStringIDAt(StartOffset + (i * 32) + 4, false).ToString(), map.ReadStringIDAt(StartOffset + (i * 32) + 8, false).ToString()),
                    StartOffset + (i * 32) + 20));
            }

            return tb;
        }

        public static TagBlock ReadEFFE(MapStream map)
        {
            int eventCount = map.ReadReflexiveAt((int)map.Position + 20);
            int eventStartOffset = (int)map.Position;

            TagBlock tb = new TagBlock();
            tb.Functions = new List<FunctionInfo>();
            for (int e = 0; e < eventCount; e++)
            {
                int eventChunkStartOffset = eventStartOffset + (e * 56);

                int beamCount = map.ReadReflexiveAt(eventChunkStartOffset + 32);
                int beamStartOffset = (int)map.Position;
                for (int b = 0; b < beamCount; b++)
                {
                    int beamChunkStartOffset = beamStartOffset + (b * 56);
                    tb.Functions.Add(new FunctionInfo(map, string.Format("Event[{0}] Beam[{1}] Color", e, b), beamChunkStartOffset + 12));
                    tb.Functions.Add(new FunctionInfo(map, string.Format("Event[{0}] Beam[{1}] Alpha", e, b), beamChunkStartOffset + 20));
                    tb.Functions.Add(new FunctionInfo(map, string.Format("Event[{0}] Beam[{1}] Width", e, b), beamChunkStartOffset + 28));
                    tb.Functions.Add(new FunctionInfo(map, string.Format("Event[{0}] Beam[{1}] Length", e, b), beamChunkStartOffset + 36));
                    tb.Functions.Add(new FunctionInfo(map, string.Format("Event[{0}] Beam[{1}] Yaw", e, b), beamChunkStartOffset + 44));
                    tb.Functions.Add(new FunctionInfo(map, string.Format("Event[{0}] Beam[{1}] Pitch", e, b), beamChunkStartOffset + 52));
                }

                int particleSystemCount = map.ReadReflexiveAt(eventChunkStartOffset + 48);
                int particleSystemStartOffset = (int)map.Position;
                for (int p = 0; p < particleSystemCount; p++)
                {
                    int particleSystemChunkStartOffset = particleSystemStartOffset + (p * 56);

                    int emitterCount = map.ReadReflexiveAt(particleSystemChunkStartOffset + 48);
                    int emitterStartOffset = (int)map.Position;
                    for (int i = 0; i < emitterCount; i++)
                    {
                        int emitterChunkStartOffset = emitterStartOffset + (i * 184);
                        tb.Functions.Add(new FunctionInfo(map, string.Format("Event[{0}] Particle System[{1}] Emitter[{2}] Emission Rate", e, p, i), emitterChunkStartOffset + 16));
                        tb.Functions.Add(new FunctionInfo(map, string.Format("Event[{0}] Particle System[{1}] Emitter[{2}] Lifespan", e, p, i), emitterChunkStartOffset + 32));
                        tb.Functions.Add(new FunctionInfo(map, string.Format("Event[{0}] Particle System[{1}] Emitter[{2}] Velocity", e, p, i), emitterChunkStartOffset + 48));
                        tb.Functions.Add(new FunctionInfo(map, string.Format("Event[{0}] Particle System[{1}] Emitter[{2}] Angular Velocity", e, p, i), emitterChunkStartOffset + 64));
                        tb.Functions.Add(new FunctionInfo(map, string.Format("Event[{0}] Particle System[{1}] Emitter[{2}] Size", e, p, i), emitterChunkStartOffset + 80));
                        tb.Functions.Add(new FunctionInfo(map, string.Format("Event[{0}] Particle System[{1}] Emitter[{2}] Tint", e, p, i), emitterChunkStartOffset + 96));
                        tb.Functions.Add(new FunctionInfo(map, string.Format("Event[{0}] Particle System[{1}] Emitter[{2}] Alpha", e, p, i), emitterChunkStartOffset + 112));
                        tb.Functions.Add(new FunctionInfo(map, string.Format("Event[{0}] Particle System[{1}] Emitter[{2}] Emission Radius", e, p, i), emitterChunkStartOffset + 132));
                        tb.Functions.Add(new FunctionInfo(map, string.Format("Event[{0}] Particle System[{1}] Emitter[{2}] Emission Angle", e, p, i), emitterChunkStartOffset + 148));
                    }
                }
            }

            return tb;
        }

        public static TagBlock ReadJPT(MapStream map)
        {
            int playerResponseCount = map.ReadReflexiveAt((int)map.Position + 108);
            int playerResponseStartOffset = (int)map.Position;

            TagBlock tb = new TagBlock();
            tb.Functions = new List<FunctionInfo>();
            for (int i = 0; i < playerResponseCount; i++)
            {
                int playerResponseChunkStartOffset = playerResponseStartOffset + (i * 76);
                tb.Functions.Add(new FunctionInfo(map, string.Format("Player Responses[{0}] Low Frequency Vibration", i), playerResponseChunkStartOffset + 40));
                tb.Functions.Add(new FunctionInfo(map, string.Format("Player Responses[{0}] High Frequency Vibration", i), playerResponseChunkStartOffset + 52));
            }

            return tb;
        }

        public static TagBlock ReadLENS(MapStream map)
        {
            int tagStartOffset = (int)map.Position;

            TagBlock tb = new TagBlock();
            tb.Functions = new List<FunctionInfo>();

            int brightnessCount = map.ReadReflexiveAt(tagStartOffset + 76);
            int brightnessStartOffset = (int)map.Position;
            for (int i = 0; i < brightnessCount; i++)
            {
                int brightnessChunkStartOffset = brightnessStartOffset + (i * 188);
                tb.Functions.Add(new FunctionInfo(map, string.Format("Brightness [{0}]", i), brightnessChunkStartOffset + 0));
            }

            int colorCount = map.ReadReflexiveAt(tagStartOffset + 84);
            int colorStartOffset = (int)map.Position;
            for (int i = 0; i < colorCount; i++)
            {
                int colorChunkStartOffset = colorStartOffset + (i * 188);
                tb.Functions.Add(new FunctionInfo(map, string.Format("Color [{0}]", i), colorChunkStartOffset + 0));
            }

            int rotationCount = map.ReadReflexiveAt(tagStartOffset + 92);
            int rotationStartOffset = (int)map.Position;
            for (int i = 0; i < rotationCount; i++)
            {
                int rotationChunkStartOffset = rotationStartOffset + (i * 188);
                tb.Functions.Add(new FunctionInfo(map, string.Format("Gel [{0}] A", i), rotationChunkStartOffset + 0));
                tb.Functions.Add(new FunctionInfo(map, string.Format("Gel [{0}] B", i), rotationChunkStartOffset + 8));
            }

            return tb;
        }

        public static TagBlock ReadLIGH(MapStream map)
        {
            int tagStartOffset = (int)map.Position;

            TagBlock tb = new TagBlock();
            tb.Functions = new List<FunctionInfo>();

            int brightnessAnimationCount = map.ReadReflexiveAt(tagStartOffset + 196);
            int brightnessAnimationStartOffset = (int)map.Position;
            for (int i = 0; i < brightnessAnimationCount; i++)
            {
                int brightnessAnimationChunkStartOffset = brightnessAnimationStartOffset + (i * 76);
                tb.Functions.Add(new FunctionInfo(map, string.Format("Brightness Animation[{0}]", i), brightnessAnimationChunkStartOffset + 0));
            }

            int colorAnimationCount = map.ReadReflexiveAt(tagStartOffset + 204);
            int colorAnimationStartOffset = (int)map.Position;
            for (int i = 0; i < colorAnimationCount; i++)
            {
                int colorAnimationChunkStartOffset = colorAnimationStartOffset + (i * 188);
                tb.Functions.Add(new FunctionInfo(map, string.Format("Color Animation[{0}]", i), colorAnimationChunkStartOffset + 0));
            }

            int gelAnimationCount = map.ReadReflexiveAt(tagStartOffset + 212);
            int gelAnimationStartOffset = (int)map.Position;
            for (int i = 0; i < gelAnimationCount; i++)
            {
                int gelAnimationChunkStartOffset = gelAnimationStartOffset + (i * 16);
                tb.Functions.Add(new FunctionInfo(map, string.Format("Gel Animation[{0}] A", i), gelAnimationChunkStartOffset + 0));
                tb.Functions.Add(new FunctionInfo(map, string.Format("Gel Animation[{0}] B", i), gelAnimationChunkStartOffset + 8));
            }

            return tb;
        }

        public static TagBlock ReadMGS2(MapStream map)
        {
            int volumeCount = map.ReadReflexiveAt((int)map.Position + 8);
            int StartOffset = (int)map.Position;

            TagBlock tb = new TagBlock();
            tb.Functions = new List<FunctionInfo>();
            for (int v = 0; v < volumeCount; v++)
            {
                int volumeStartOffset = StartOffset + (v * 152);
                tb.Functions.Add(new FunctionInfo(map, string.Format("Volumes[{0}] Offset", v), volumeStartOffset + 16));
                tb.Functions.Add(new FunctionInfo(map, string.Format("Volumes[{0}] Radius", v), volumeStartOffset + 24));
                tb.Functions.Add(new FunctionInfo(map, string.Format("Volumes[{0}] Brightness", v), volumeStartOffset + 32));
                tb.Functions.Add(new FunctionInfo(map, string.Format("Volumes[{0}] Color", v), volumeStartOffset + 40));
                tb.Functions.Add(new FunctionInfo(map, string.Format("Volumes[{0}] Facing", v), volumeStartOffset + 48));

                int aspectCount = map.ReadReflexiveAt(volumeStartOffset + 56);
                int aspectStartOffset = (int)map.Position;
                for (int a = 0; a < aspectCount; a++)
                {
                    int aspectChunkStartOffset = aspectStartOffset + (a * 28);
                    tb.Functions.Add(new FunctionInfo(map, string.Format("Volumes[{0}] Aspect[{1}] Along-Axis Scale", v, a), aspectChunkStartOffset + 0));
                    tb.Functions.Add(new FunctionInfo(map, string.Format("Volumes[{0}] Aspect[{1}] Away-From-Axis Scale", v, a), aspectChunkStartOffset + 8));
                }
            }

            return tb;
        }

        public static TagBlock ReadPrt3(MapStream map)
        {
            int StartOffset = (int)map.Position;

            TagBlock tb = new TagBlock();
            tb.Functions = new List<FunctionInfo>();
            tb.Functions.Add(new FunctionInfo(map, "Color", StartOffset + 36));
            tb.Functions.Add(new FunctionInfo(map, "Alpha", StartOffset + 52));
            tb.Functions.Add(new FunctionInfo(map, "Scale", StartOffset + 68));
            tb.Functions.Add(new FunctionInfo(map, "Rotation", StartOffset + 84));
            tb.Functions.Add(new FunctionInfo(map, "Frame Index", StartOffset + 100));

            return tb;
        }

        public static TagBlock ReadPRTM(MapStream map)
        {
            int StartOffset = (int)map.Position;

            TagBlock tb = new TagBlock();
            tb.Functions = new List<FunctionInfo>();
            tb.Functions.Add(new FunctionInfo(map, "Scale X", StartOffset + 40));
            tb.Functions.Add(new FunctionInfo(map, "Scale Y", StartOffset + 56));
            tb.Functions.Add(new FunctionInfo(map, "Scale Z", StartOffset + 72));
            tb.Functions.Add(new FunctionInfo(map, "Frame Index", StartOffset + 88));

            return tb;
        }

        public static TagBlock ReadTDTL(MapStream map)
        {
            int arcCount = map.ReadReflexiveAt((int)map.Position + 104);
            int StartOffset = (int)map.Position;

            TagBlock tb = new TagBlock();
            tb.Functions = new List<FunctionInfo>();
            for (int a = 0; a < arcCount; a++)
            {
                int arcStartOffset = StartOffset + (a * 324);
                tb.Functions.Add(new FunctionInfo(map, string.Format("Arc[{0}] Horizontal Range", a), arcStartOffset + 52));
                tb.Functions.Add(new FunctionInfo(map, string.Format("Arc[{0}] Vertical Range", a), arcStartOffset + 60));
                tb.Functions.Add(new FunctionInfo(map, string.Format("Arc[{0}] Roughness", a), arcStartOffset + 72));
                tb.Functions.Add(new FunctionInfo(map, string.Format("Arc[{0}] Range Collision Scale", a), arcStartOffset + 220));
                tb.Functions.Add(new FunctionInfo(map, string.Format("Arc[{0}] Bightness Collision Scale", a), arcStartOffset + 228));

                int coreCount = map.ReadReflexiveAt(arcStartOffset + 212);
                int coreStartOffset = (int)map.Position;
                for (int c = 0; c < coreCount; c++)
                {
                    int coreChunkStartOffset = coreStartOffset + (c * 56);
                    tb.Functions.Add(new FunctionInfo(map, string.Format("Arc[{0}] Core[{1}] Thickness", a, c), coreChunkStartOffset + 16));
                    tb.Functions.Add(new FunctionInfo(map, string.Format("Arc[{0}] Core[{1}] Color", a, c), coreChunkStartOffset + 24));
                    tb.Functions.Add(new FunctionInfo(map, string.Format("Arc[{0}] Core[{1}] Brightness/Time", a, c), coreChunkStartOffset + 32));
                    tb.Functions.Add(new FunctionInfo(map, string.Format("Arc[{0}] Core[{1}] Brightness/Facing", a, c), coreChunkStartOffset + 40));
                    tb.Functions.Add(new FunctionInfo(map, string.Format("Arc[{0}] Core[{1}] Along-Axis Scale", a, c), coreChunkStartOffset + 48));
                }
            }

            return tb;
        }
    }
}
